/*
 * conns.c - These routines handle the linked list of active connections.
 */
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "config.h"
#include "log.h"
#include "utils.h"
#include "conns.h"
#include "error.h"

/*
 * Add a new connection to the linked list
 */
struct conn_s *new_conn(enum conn_type conntype, int fd)
{
	extern struct conn_s connections;

	struct conn_s *connptr = &connections;
	struct conn_s *connptr2;

	/* Make sure the socket does exist and the connect type is correct */
	if (fd < 0 || (conntype != CONNECTINGCONN && conntype != CLIENTCONN)) {
		log("new_conn: problem creating connection (%d, %d)",
		    conntype, fd);
		tinyerr = ERR_INVALID;
		return NULL;
	}

	for (; connptr->next; connptr = connptr->next);;

	connptr2 = connptr;
	if (!(connptr2->next = connptr = xmalloc(sizeof(struct conn_s)))) {
		log("new_conn: could not allocate memory for conn");
		tinyerr = ERR_NOMEM;
		return NULL;
	}

	/* Hold the buffer inside the connection struct */
	if (!(connptr->buffer = xmalloc(BUFFER))) {
		log("new_conn: could not allocate memory for buffer");
		free(connptr);
		connptr2->next = NULL;
		tinyerr = ERR_NOMEM;
		return NULL;
	}
	connptr->buffsize = 0;

	connptr->fd = fd;
	connptr->host = connptr->request = NULL;
	connptr->conntype = conntype;
	connptr->next = NULL;
	connptr->prev = connptr2;
	connptr->other = NULL;
	connptr->inittime = time(NULL);
	connptr->actiontime = time(NULL);
	connptr->killme = FALSE;
#ifdef ANON
	if (conntype == CONNECTINGCONN) {
		connptr->postheader = TRUE;
	} else {
		connptr->postheader = FALSE;
	}
#endif
	connptr->port = HTTP_PORT;

	stats.num_cons++;

	return connptr;
}

/*
 * Delete a connection from the linked list
 */
int del_conn(struct conn_s *delconn)
{
	if (!delconn) {
		log("internal error: del_conn is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

	delconn->prev->next = delconn->next;
	if (delconn->next)
		delconn->next->prev = delconn->prev;
	free(delconn);
	return 0;
}

/*
 * Flag a connection for deletion on the next garbage collect
 */
int predel_conn(struct conn_s *delconn)
{
	if (!delconn) {
		log("internal error: predel_conn is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

	close(delconn->fd);
	if (delconn->host && delconn->request)
		stats.num_reqs++;
	if (delconn->host)
		free(delconn->host);
	if (delconn->request)
		free(delconn->request);
	if (delconn->buffer)
		free(delconn->buffer);

	delconn->killme = TRUE;
	delconn->conntype = PLACEHOLDER;
	return 0;
}

/*
 * Check for connections that have been idle too long
 */
void conncoll(void)
{
	struct conn_s  *connptr = &connections;

	while (connptr) {
		if (((time(NULL) - connptr->actiontime) > STALECONN_TIME)
		    && (connptr->conntype != PLACEHOLDER)) {
			predel_conn(connptr);
			stats.num_idles++;
			connptr = &connections;
		} else {
			connptr = connptr->next;
		}
	}
}

/*
 * Actually remove all entries in the linked list that have been marked for
 * deletion.
 */
void garbcoll(void)
{
	struct conn_s  *connptr = &connections;

	stats.num_garbage++;

	while (connptr) {
		if (connptr->killme == TRUE) {
			del_conn(connptr);
			connptr = &connections;
		} else {
			connptr = connptr->next;
		}
	}
}
