/*
 * reqs.c - Handle relaying between connecitons and establishing of
 * connections to servers.
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sysexits.h>

#include "config.h"
#include "tinyproxy.h"
#include "sock.h"
#include "utils.h"
#include "conns.h"
#include "log.h"
#include "reqs.h"
#include "error.h"


/*
 * Reads the bytes from the socket, and adds them to the buffer.
 * Takes a connection and returns the number of bytes read.
 */
static int readbuff(struct conn_s *connptr)
{
	int bytesin;
	char c;

	if (!connptr) {
		log("readbuff: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

	/*
	 * Check to see if the connection was closed. If it was, but we haven't
	 * finished sending, just return. We'll handle the closed socket when
	 * the all the data in the buffer is sent.
	 */
	bytesin = recv(connptr->fd, &c, sizeof(c), MSG_PEEK);
	if ((bytesin == 0) && (connptr->buffsize > 0)) {
		return 0;
	}

	/* The general case */
	if (connptr->buffsize < 0 || connptr->buffsize > BUFFER) {
		log("readbuff: invalid buffsize (%d)", connptr->buffsize);
		tinyerr = ERR_INVALID;
		return -1;
	}
	bytesin = recv(connptr->fd, connptr->buffer + connptr->buffsize,
		       BUFFER - connptr->buffsize, 0);

	if (bytesin > 0) {
		connptr->buffsize += bytesin;
		return bytesin;
	} else if (bytesin == 0) {
		/* connection was closed by client */
		tinyerr = ERR_CONNCLOSED;
		return -1;
	} else {
		switch (errno) {
		case EWOULDBLOCK:
		case EINTR:
			tinyerr = ERR_RETRY;
			return -1;
		default:
			log("readbuff: recv (%s)", strerror(errno));
			tinyerr = ERR_CONNERR;
			return -1;
		}
	}
}

/*
 * Write the bytes in the buffer to the socket.
 * Takes a connection and returns the number of bytes written.
 */
static int writebuff(struct conn_s *connptr)
{
	int bytessent;

	if (!connptr) {
		log("writebuff: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

	if (connptr->buffsize < 0 || connptr->buffsize > BUFFER) {
		log("writebuff: invalid buffsize (%d)", connptr->buffsize);
		tinyerr = ERR_INVALID;
		return -1;
	}
	bytessent = send(connptr->other->fd, connptr->buffer,
			 connptr->buffsize, 0);

	if (bytessent > 0) {
		/* bytes sent, adjust buffer */
		memmove(connptr->buffer, connptr->buffer + bytessent,
			connptr->buffsize - bytessent);
		connptr->buffsize -= bytessent;
		return bytessent;
	} else if (bytessent == 0) {
		return 0;
	} else {
		switch (errno) {
		case EWOULDBLOCK:
		case ENOBUFS:
			tinyerr = ERR_RETRY;
			return -1;
		case EPIPE:
			tinyerr = ERR_CONNCLOSED;
			return -1;
		default:
			log("writebuff: send (%s)", strerror(errno));
			tinyerr = ERR_CONNERR;
			return -1;
		}
	}
}

#ifdef XTINYPROXY
static void Add_XTinyproxy_Header(struct conn_s *connptr)
{
	struct sockaddr_in peername;
	int peerlen = sizeof(struct sockaddr_in);
	char header_line[128];
	unsigned char* end_buffer;
	int length;
	
	getpeername(connptr->fd, (struct sockaddr*)&peername, &peerlen);
	length = sprintf(header_line, "X-Tinyproxy: %s\r\n", inet_ntoa(*(struct in_addr*)&peername.sin_addr.s_addr));
	
	/* Add the new header line to the end of the buffer */
	end_buffer = strstr(connptr->buffer, "\r\n") + 2;
	
	/* shift the buffer down so we can insert the new header line */
	memmove((end_buffer + length), end_buffer, (connptr->buffsize - (end_buffer - connptr->buffer)));
	memcpy(end_buffer, header_line, length);
	connptr->buffsize += length;
}
#endif

/*
 * Parse a client HTTP request and then establish connection.
 */
static int clientreq(struct conn_s *connptr)
{
#ifndef UPSTREAM
	unsigned char *line_end, *location, *page, *port, *method;
	unsigned char inbuf[BUFFER];
#endif
	int bytesin;
	int fd;

	struct sockaddr_in name;
	int namelen = sizeof(name);
	char *socketname;
	struct hostent *peername;

	if (!connptr) {
		log("clientreq: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}
	
	/* Read in the requests for the connection */
	if ((bytesin = readbuff(connptr)) < 0) {
		if (tinyerr == ERR_RETRY) {
			return 0;
		} else {
			predel_conn(connptr);
			if (tinyerr == ERR_CONNCLOSED)
				return 0;

			log("clientreq: connection error");
			return -1;
		}
	}
	stats.num_rx += bytesin;

	/* Log the incoming connection */
	if (getpeername(connptr->fd, (struct sockaddr*)&name, &namelen) != 0) {
		log("Incoming connection: could not determine peer address");
	} else {
		socketname = inet_ntoa(name.sin_addr);
		if ((peername = gethostbyaddr((char*)&name.sin_addr.s_addr, sizeof(name.sin_addr.s_addr), AF_INET)) == NULL) {
			log("Incoming connection: [%s]", socketname);
		} else {
			log("Incoming connection: %s [%s]", peername->h_name, socketname);
		}
	}

#ifdef XTINYPROXY
	/* Add a X-Tinyproxy header which contains our IP address */
	Add_XTinyproxy_Header(connptr);
#endif

#ifndef UPSTREAM	
	/*
	 * Find the end of the first line, copy it into our buffer here and
	 * then delete it from the connection buffer (don't want to send it
	 * twice.)
	 */
	if (!(line_end = strstr(connptr->buffer, "\r\n"))) {
		httperr(connptr->fd, 400, "Incomplete request");
		predel_conn(connptr);
		tinyerr = ERR_REQERR;
		return -1;
	}
	line_end += 2;
	memcpy(inbuf, connptr->buffer, line_end - connptr->buffer);
	inbuf[line_end - connptr->buffer] = '\0';

	/* Shift the buffer up by the len of the request */
	memmove(connptr->buffer, line_end,
		connptr->buffsize - (line_end - connptr->buffer));
	connptr->buffsize -= (line_end - connptr->buffer);
	
	/* Build a new request from the first line of the header */
	if (!(connptr->request = xmalloc(strlen(inbuf)))) {
		log("clientreq: cannot allocate buffer for request");
		httperr(connptr->fd, 503, "Internal Error");
		predel_conn(connptr);
		return -1;
	}
	if (!(location = index(inbuf, ' '))) {
		httperr(connptr->fd, 400, "Unrecognizable request");
		predel_conn(connptr);
		tinyerr = ERR_REQERR;
		return -1;
	}
	*(location++) = '\0';
	strncpy(connptr->request, inbuf, strlen(inbuf) + 1);
	strncat(connptr->request, " /", 3);
	
	if ((unsigned char*)strstr(location, "http://") != location) {
		httperr(connptr->fd, 400, "Oops. Not a HTTP location");
		predel_conn(connptr);
		tinyerr = ERR_REQERR;
		return -1;
	}
		
	location += 7;

	page = strpbrk(location, "/ ");
	if (!page) {
		httperr(connptr->fd, 400, "Incomplete URL");
		predel_conn(connptr);
		tinyerr = ERR_REQERR;
		return -1;
	} else {
		if (*page == '/') {
			*(page++) = '\0';
			if (!(method = index(page, ' '))) {
				httperr(connptr->fd, 400, "Not a valid method");
				predel_conn(connptr);
				tinyerr = ERR_REQERR;
				return -1;
			}
			*(method++) = '\0';
		} else {
			method = page + 1;
			*(page) = '\0';
			page = "";
		}
	}	    

	if ((port = index(location, ':')) != NULL) {
		*(port++) = '\0';
		connptr->port = atoi(port);
	} else {
		connptr->port = 80;
	}

	if (!(connptr->host = xstrdup(location))) {
		log("clientreq: cannot duplicate string");
		httperr(connptr->fd, 503, "Internal Error");
		predel_conn(connptr);
		return -1;
	}
	
	strncat(connptr->request, page, strlen(page) + 1);
	strncat(connptr->request, " ", 2);
	strncat(connptr->request, method, strlen(method) + 1);

	if((strlen(config.stathost) > 0) && 
	   !strncmp(connptr->host, config.stathost, strlen(config.stathost))) {
		showstats(connptr->fd);
		predel_conn(connptr);
		return 0;
	}

	if ((fd = opensock(connptr->host, connptr->port)) < 0) {
#else
	/* This is if we're using the upstream proxy */ 
	if ((fd = opensock(UPSTREAM_PROXY_NAME, UPSTREAM_PROXY_PORT)) < 0) {
#endif
		if (tinyerr == ERR_RETRY)
			log("clientreq: retry error from opensock");
		else
			log("clientreq: cannot opensock");
		httperr(connptr->fd, 500, "Unable to connect to host (cannot create sock)");
		predel_conn(connptr);
		stats.num_badcons++;
		return -1;
	}
	
	if (!(connptr->other = new_conn(CONNECTINGCONN, fd))) {
		log("clientreq: could not create connection");
		httperr(connptr->fd, 404, "Unable to connect to host");
		predel_conn(connptr);
		stats.num_badcons++;
		return -1;
	}
	connptr->other->other = connptr;
	connptr->conntype = PLACEHOLDER;
	return 0;
}

/*
 * Finish the client request
 */
static int clientreq_finish (struct conn_s *connptr)
{
#ifndef UPSTREAM
	int bytesout;
#endif
	int sockopt, len = sizeof(sockopt);

	getsockopt(connptr->fd, SOL_SOCKET, SO_ERROR, &sockopt, &len);
	if (sockopt != 0) {
		log("clientreq_finish: could not create connection (%s)", strerror(errno));
		return -1;
	}
	
	connptr->conntype = SERVERCONN;
	connptr->other->conntype = IGNORECONN;
#ifndef UPSTREAM
	bytesout = send(connptr->fd, connptr->other->request, strlen(connptr->other->request), 0);
	stats.num_tx += bytesout;
#endif
	return 0;
}

/*
 * Read in the bytes from the socket
 */
static int readreq(struct conn_s *connptr)
{
	int bytesin;

	if (!connptr) {
		log("readreq: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

	if ((bytesin = readbuff(connptr)) < 0) {
		if (tinyerr == ERR_RETRY) {
			return 0;
		} else if (tinyerr == ERR_CONNCLOSED
			   && connptr->buffsize > 0) {
			return 0;
		} else {
			predel_conn(connptr);
			predel_conn(connptr->other);
			return -1;
		}
	}

	stats.num_rx += bytesin;
	connptr->actiontime = connptr->other->actiontime = time(NULL);
	return 0;
}

#ifdef ANON
/*
 * Anonymouns Headers
 */
static int anonheader(struct conn_s *connptr)
{
	extern struct allowedhdr_s *allowedhdrs;

	struct allowedhdr_s *allowedptr = allowedhdrs;
	unsigned char *newbuff;
	unsigned char *begin, *end, *buffer_end;
	unsigned int len, offset = 0;

	if (!connptr) {
		log("anonheader: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}
	if (!allowedhdrs) {
		log("anonheader: allowedhdrs is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}
	
	/*
	 * Convert the connection buffer into a stack of header lines
	 */
	if (!(newbuff = xmalloc(BUFFER))) {
		log("anonheader: cannnot allocate buffer");
		return -1;
	}
	offset = 0;

	buffer_end = connptr->buffer + connptr->buffsize;

	begin = connptr->buffer;
	while (((end = strstr(begin, "\r\n")) != begin) && (begin < buffer_end)) {
		end += 2;
		len = end - begin;
		
		/*
		 * Loop through the anonymous headers and see if this line
		 * matches
		 */
		for (allowedptr = allowedhdrs;
		     allowedptr;
		     allowedptr = allowedptr->next) {
			if (!strncmp(begin, allowedptr->hdrname,
				     strlen(allowedptr->hdrname))) {
				/*
				 * The line matches, so add it to the new
				 * buffer
				 */
				memcpy(newbuff + offset, begin, len);
				offset += len;
				break;
			}
		}
		begin = end;
	}

	/*
	 * We've cleansed the headers, now replace the connection buffer with
	 * our new header
	 */
	memcpy(newbuff + offset, "\r\n", 2);
	free(connptr->buffer);
	connptr->buffer = newbuff;
	connptr->buffsize = offset + 2;
	return 0;
}
#endif

/*
 * Write the bytes from the buffer to the socket
 */
static int writereq(struct conn_s *connptr)
{
	int bytessent;

	if (!connptr) {
		log("writereq: connptr is NULL");
		tinyerr = ERR_INVALID;
		return -1;
	}

#ifdef ANON
	if ((config.anonymous == TRUE) && (connptr->postheader == FALSE)) {
		/*
		 * Unless we have the entire go back and get the rest.
		 * Note: this method limits us to a header which can be no
		 * larger than the maximum size of the buffer (currently 8K).
		 */
		if (!strstr(connptr->buffer, "\r\n\r\n"))
			return 0;
		else
			connptr->postheader = TRUE;

		anonheader(connptr);
	}
#endif

	if ((bytessent = writebuff(connptr)) < 0) {
		if (tinyerr == ERR_RETRY) {
			return 0;
		} else {
			predel_conn(connptr);
			predel_conn(connptr->other);
			return -1;
		}
	}
	stats.num_tx += bytessent;
	return 0;
}

/*
 * Loop that checks for new connections, dispatches to the correct
 * routines if bytes are pending, checks to see if it's time for a
 * garbage collect.
 */
int getreqs(void)
{
	extern int setup_fd;
	extern struct conn_s connections;

	static int garbc = 0;
	fd_set readfds, writefds;
	struct conn_s *connptr;
	int fd, maxfd;

	if (setup_fd <= 0) {
		log("getreqs: setup_fd not a socket");
		tinyerr = ERR_INVALID;
		return -1;
	}

	/* Wait for the any activity on the sockets */
	if (++garbc >= GARBCOLL_INTERVAL) {
		garbcoll();
		conncoll();
		garbc = 0;
	}

	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_SET(setup_fd, &readfds);

	maxfd = setup_fd;
	connptr = &connections;
	while (connptr) {
		if (connptr->conntype != PLACEHOLDER) {
			if (BUFFER - connptr->buffsize > 0)
				FD_SET(connptr->fd, &readfds);
			if (connptr->buffsize > 0)
				FD_SET(connptr->fd, &writefds);
			if (connptr->fd > maxfd)
				maxfd = connptr->fd;
		}
		if (connptr->conntype == CONNECTINGCONN) {
			FD_SET(connptr->fd, &writefds);
		}
		connptr = connptr->next;
	}
  	   
	if (select(FD_SETSIZE, &readfds, &writefds, NULL, NULL) < 0)
		return 0;

	/* Check to see if there are new connections pending */
	if (FD_ISSET(setup_fd, &readfds)) {
		if ((fd = listen_sock()) >= 0) {
			if((!config.cutoffload) || (load < config.cutoffload)) {
				if (!new_conn(CLIENTCONN, fd)) {
					log("getreqs: cannnot make new connection");
					stats.num_refused++;
					httperr(fd, 503, "can not make a new connection");
					close(fd);
				}
			} else {
				stats.num_refused++;
				httperr(fd, 503, "tinyproxy is not accepting connections due to high system load");
				close(fd);
			}
		} else {
			log("getreqs: accept error");
		}
	}

	/* 
	 * Loop through the connections and dispatch them to the appropriate
	 * handler
	 */
	for (connptr = &connections; connptr; connptr = connptr->next) {
		switch(connptr->conntype) {
		case CLIENTCONN:
			if (FD_ISSET(connptr->fd, &readfds))
				clientreq(connptr);
			break;

		case CONNECTINGCONN:
			if (FD_ISSET(connptr->fd, &writefds))
			    clientreq_finish(connptr);
			break;

		case SERVERCONN:
		case IGNORECONN:
			if (FD_ISSET(connptr->fd, &readfds))
				readreq(connptr);
			if (FD_ISSET(connptr->fd, &writefds))
				writereq(connptr);
			break;

		default:
			break;
		}
	}
	return 0;
}
