/*
 * tinyproxy.c - Main tinyproxy source.
 */
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sysexits.h>
#include <pwd.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>

#include "config.h"
#include "global.h"
#include "tinyproxy.h"
#include "utils.h"
#include "log.h"
#include "sock.h"
#include "conns.h"
#include "reqs.h"
#include "error.h"

/*
 * External definition for daemon syscall from BSD library
 */
extern int daemon(int nochdir, int noclose);

/*
 * Display an error to the client.
 */
int httperr(int fd, int err, char *msg)
{
	char outbuf[BUFFER];
	static char *premsg = \
"HTTP/1.0 %d %s\r\n\
Content-type: text/html\r\n\r\n\
<html><head><title>%s</title></head>\r\n\
<body>\r\n\
<font size=\"+2\">Cache Error!</font><br>\r\n\
An error of type %d occurred: %s\r\n\
<hr>\r\n\
<font size=\"-1\"><em>Generated by %s</em></font>\r\n\
</body></html>\r\n";

	if (fd <= 0 || err <= 0 || !msg) {
		log("httperr: argument error (%d, %d, %x)", fd, err, msg);
		tinyerr = ERR_INVALID;
		return -1;
	}

	sprintf(outbuf, premsg, err, msg, msg, err, msg, VERSION);
	send(fd, outbuf, strlen(outbuf), 0);
	return 0;
}

/*
 * Dump info to the logfile
 */
static void dumpdebug(void)
{
	extern struct conn_s connections;

	struct conn_s *connptr = &connections;
	int clients = 0, servers = 0, holders = 0, ignores = 0;

	log("SIGUSR1 received, debug dump follows.");

	while (connptr) {
		switch (connptr->conntype) {
		case CLIENTCONN:
			clients++;
			break;
		case SERVERCONN:
			servers++;
			break;
		case PLACEHOLDER:
			holders++;
			break;
		case IGNORECONN:
			ignores++;
			break;
		default:
			break;
		}
		connptr = connptr->next;
	}
	log("servers: %d, clients: %d, ignores: %d, holders: %d",
	    servers, clients, ignores, holders);
	log("total requests handled: %lu", stats.num_reqs);
	log("total connections handled: %lu", stats.num_cons);
	log("total sockets listened: %lu", stats.num_listens);
	log("total sockets opened: %lu", stats.num_opens);
	log("total bad opens: %lu", stats.num_badcons);
	log("total bytes tx: %lu", stats.num_tx);
	log("total bytes rx: %lu", stats.num_rx);
	log("connections refused due to load: %lu", stats.num_refused);
	log("garbage collections: %lu", stats.num_garbage);
	log("idle connections killed: %lu", stats.num_idles);
	log("end debug dump.");
}

/*
 * Handle a signal
 */
static void takesig(int sig)
{
	switch (sig) {
	case SIGUSR1:
		dumpdebug();
		break;
	case SIGHUP:
		log("SIGHUP received, cleaning up...");
		garbcoll();
		conncoll();
		log("Finished cleaning memory/connections.");
		break;
	case SIGTERM:
		config.quit = TRUE;
		break;
	case SIGALRM:
		calcload();
		alarm(LOAD_RECALCTIMER);
		break;
	}
	if (sig != SIGTERM)
		signal(sig, takesig);
}

/*
 * Display usage to the user on stderr.
 */
static void usagedisp(void)
{
	fprintf(stderr, "%s\n", VERSION);
	fprintf(stderr, "this software is GPLed; see the file 'COPYING' included with tinyproxy.\n");
	fprintf(stderr, "mail comments/questions/money to: sdyoung@well.com.\n");
	fprintf(stderr, "usage: tinyproxy [args]\n");
	fprintf(stderr, "possible values for [args]:\n");
	fprintf(stderr, "\t-v\t\tdisplay version.\n");
	fprintf(stderr, "\t-h\t\tdisplay this.\n");
	fprintf(stderr, "\t-d\t\tdon't daemonize.\n");
	fprintf(stderr, "\t-p [port]\tlisten on [port].\n");
	fprintf(stderr, "\t-l [logfile]\tlog to [logfile].\n");
	fprintf(stderr, "\t-S\t\tlog using the syslog instead.\n");
	fprintf(stderr, "\t-w [load]\tstop accepting new conns at [load].  0 disables.\n");
	fprintf(stderr, "\t-s [stathost]\tset stathost to [stathost].\n");
	fprintf(stderr, "\t-u [user]\tchange to [user] after startup.  \"\" disables.\n");
#ifdef ANON
	fprintf(stderr, "\t-a [header]\tallow [header] through the anon block.\n");
#endif
	fprintf(stderr, "default port: %d\n", DEFAULT_PORT);
	fprintf(stderr, "default logfile: %s\n", DEFAULT_LOG);
#ifdef ANON
	fprintf(stderr, "anon mode is compiled in.\n");
#endif
}

int main(int argc, char **argv)
{
	char optch;
	extern char *optarg;
	flag usage = FALSE, godaemon = TRUE, changeid = FALSE;
	struct passwd *thisuser = NULL;
#ifdef ANON
	extern struct allowedhdr_s *allowedhdrs;
	struct allowedhdr_s *allowedptr = NULL;
#endif

	while ((optch = getopt(argc, argv, "vhdp:l:Sa:w:s:u:")) != EOF) {
		switch (optch) {
		case 'v':
			fprintf(stderr, "%s\n", VERSION);
			exit(EX_OK);
		case 'h':
		case '?':
			usage = TRUE;
			break;
		case 'p':
			if (!(config.port = atoi(optarg))) {
				log("bad port on commandline, defaulting to %d", DEFAULT_PORT);
				config.port = DEFAULT_PORT;
			}
			break;
		case 'l':
			if (!(config.logf_name = xstrdup(optarg))) {
				log("bad log file, defaulting to %s", DEFAULT_LOG);
				config.logf_name = DEFAULT_LOG;
			}
			break;
		case 'S': /* Use the syslog function to handle logging */
			config.syslog = TRUE;
			break;
		case 'd':
			godaemon = FALSE;
			break;
		case 'w':
			sscanf(optarg, "%f", &config.cutoffload);
			break;
		case 's':
			if (!(config.stathost=xstrdup(optarg))) {
				log("bad stathost, defaulting to %s", DEFAULT_STATHOST);
				config.stathost = DEFAULT_STATHOST;
			}
			break;
		case 'u' : 
			if (!(config.changeuser=xstrdup(optarg))) {
				log("bad user name, defaulting to %s", DEFAULT_USER);
				config.changeuser = DEFAULT_USER;
			}
			break;
#ifdef ANON
		case 'a':
			config.anonymous = TRUE;
			if (!allowedhdrs) {
				if (!(allowedhdrs = allowedptr = xmalloc(sizeof(struct allowedhdr_s)))) {
					log("tinyproxy: cannot allocate headers");
					exit(EX_SOFTWARE);
				}
			} else {
				if (!(allowedptr->next = xmalloc(sizeof(struct allowedhdr_s)))) {
					log("tinyproxy: cannot allocate next header");
					exit(EX_SOFTWARE);
				}
				allowedptr = allowedptr->next;
			}
			if (!(allowedptr->hdrname = xstrdup(optarg))) {
				log("tinyproxy: cannot duplicate string");
				exit(EX_SOFTWARE);
			}
			allowedptr->next = NULL;
#else
		case 'a':
			log("-a only applies when ANON mode has been compiled in!");
			exit(EX_USAGE);
#endif
		}
	}

	if (usage == TRUE) {
		usagedisp();
		exit(EX_OK);
	}

	/* Open the log file if not using syslog */
	if (config.syslog == FALSE) {
		if (!(config.logf = fopen(config.logf_name, "a"))) {
			fprintf(stderr, "Unable to open logfile %s for appending!", config.logf_name);
			exit(EX_CANTCREAT);
		}
	} else {
		if (godaemon == TRUE)
			openlog("tinyproxy", LOG_PID, LOG_DAEMON);
		else
			openlog("tinyproxy", LOG_PID, LOG_USER);
	}

	log("%s starting...", VERSION);

	if(strlen(config.changeuser)) {
		if((getuid() != 0) && (geteuid() != 0)) {
			log("not running as root, therefore not changing uid/gid.");
		} else {
			changeid = TRUE;
			if(!(thisuser=getpwnam(config.changeuser))) {
				log("unable to find user \"%s\"!", config.changeuser);
				exit(EX_NOUSER);
			}
			log("changing to user \"%s\" (%d/%d).",
			    config.changeuser, thisuser->pw_uid, thisuser->pw_gid);
		}
	} 

	if (godaemon == TRUE)
		daemon(0, STDERR_FILENO);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGUSR1, takesig);
	signal(SIGTERM, takesig);
	signal(SIGHUP, takesig);
	signal(SIGALRM, takesig);
	alarm(LOAD_RECALCTIMER);
	calcload();

	if (init_listen_sock(config.port) < 0) {
		log("unable to bind port %d!", config.port);
		exit(EX_UNAVAILABLE);
	}
	if(changeid == TRUE) {
		setuid(thisuser->pw_uid);
		setgid(thisuser->pw_gid);
	}
	log("now accepting connections.");

	while (config.quit == FALSE) {
		if (getreqs() < 0)
			break;
	}

	log("shutting down.");
	de_init_listen_sock();

	if (config.syslog == FALSE)
		fclose(config.logf);
	else
		closelog();

#ifdef ANON
	allowedptr = allowedhdrs;
	while (allowedptr) {
		struct allowedhdr_s *prev = allowedptr;
		free(allowedptr->hdrname);
		allowedptr = allowedptr->next;
		free(prev);
	}
#endif

	exit(EX_OK);
}
